theory Metatheory
imports Typing Semantics

begin

lemma weakening_tm:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  and "\<Gamma>\<union>\<Gamma>';\<Delta> \<Turnstile> OK"
  shows "\<Gamma>\<union>\<Gamma>';\<Delta> \<turnstile> e : \<tau>"
using prems proof (induct arbitrary: \<Gamma>' M')
  case (t_abs L \<Delta> \<Gamma> \<tau>1 \<tau>2 e \<Gamma>')
  { fix x
    assume H: "\<not> x free in e \<and> x \<notin> L \<and> \<not> (\<Gamma>\<union>\<Gamma>') defines x"
    hence "(\<Gamma>\<union> \<Gamma>'),(x:\<tau>1);\<Delta> \<Turnstile> OK" using t_abs.prems by (auto simp add: wf_env_def dom_def)
    hence "(\<Gamma>\<union>\<Gamma>'),(x:\<tau>1);\<Delta> \<turnstile> freshen_tm x e : \<tau>2" using t_abs.hyps H
      by (auto simp add: dom_def)
  }
  thus ?case using t_abs by auto
next
  case (t_poly L \<Delta> \<Gamma> \<tau> v \<Gamma>')
  { fix a
    assume H: "\<not> a free in term v \<and> \<not> a free in type \<tau> \<and> a \<notin> L \<and> 
               \<not> a free in env (\<Gamma>\<union>\<Gamma>') \<and> \<not> \<Delta> defines a"
    hence "\<Gamma>\<union>\<Gamma>';\<Delta>,(a:\<star>) \<Turnstile> OK" using t_poly.prems wf_env_ty_cons by auto      
    hence "\<Gamma>\<union>\<Gamma>';\<Delta>,(a:\<star>) \<turnstile> freshen_tmty a v : freshen_ty a \<tau>"
      using t_poly H free_ty_env_subset[of a "\<Gamma>\<union>\<Gamma>'" \<Gamma>] by (auto simp add: dom_def)
  }
  thus ?case using t_poly by auto
qed auto

lemma weakening_ty:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  and "\<Gamma>;(\<Delta>\<union>\<Delta>') \<Turnstile> OK"
  shows "\<Gamma>;(\<Delta>\<union>\<Delta>') \<turnstile> e : \<tau>"
using prems proof (induct arbitrary: \<Delta>')
  case t_tyapp thus ?case using ok_weakening_ty_union wf_env_wf_env_ty by auto
next
  case (t_abs L \<Delta> \<Gamma> \<tau>1 \<tau>2 e \<Delta>')
  have Hc1: "(\<Delta>\<union>\<Delta>') \<Turnstile> \<tau>1" using prems ok_weakening_ty_union wf_env_wf_env_ty by auto
  { fix x
    assume H: "\<not> x free in e \<and> x \<notin> L \<and> \<not> \<Gamma> defines x"
    hence "\<Gamma>,(x:\<tau>1);\<Delta>\<union>\<Delta>' \<Turnstile> OK" using t_abs.prems Hc1 by (auto simp add: wf_env_cons)
    hence "\<Gamma>,(x:\<tau>1);\<Delta>\<union>\<Delta>' \<turnstile> freshen_tm x e : \<tau>2" using t_abs.hyps H
      by (auto simp add: dom_def)
  }
  thus ?case using t_abs Hc1 by auto
next
  case (t_poly L \<Delta> \<Gamma> \<tau> v \<Delta>')
  { fix a
    assume H: "\<not> a free in term v \<and> \<not> a free in type \<tau> \<and> a \<notin> L \<and> 
               \<not> a free in env \<Gamma> \<and> \<not> (\<Delta>\<union>\<Delta>') defines a"
    hence "\<Gamma>;(\<Delta>\<union>\<Delta>'),(a:\<star>) \<Turnstile> OK" using t_poly.prems wf_env_ty_cons[of _ "\<Delta>\<union>\<Delta>'"] by auto
    hence "\<Gamma>;(\<Delta>\<union>\<Delta>'),(a:\<star>) \<turnstile> freshen_tmty a v : freshen_ty a \<tau>"
      using t_poly H by (auto simp add: dom_def)
  }
  thus ?case using t_poly by auto
qed auto

lemma weakening:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  and "(\<Gamma>\<union>\<Gamma>');(\<Delta>\<union>\<Delta>') \<Turnstile> OK"
  shows "(\<Gamma>\<union>\<Gamma>');(\<Delta>\<union>\<Delta>') \<turnstile> e : \<tau>"
proof -
  have "(\<Delta> \<union> \<Delta>') \<Turnstile> TY" using prems wf_env_wf_env_ty by auto
  hence "\<Gamma>;(\<Delta> \<union> \<Delta>') \<Turnstile> OK" 
    using ok_weakening_ty_env_gen[of \<Gamma> \<Delta>] ty_wf_env[of \<Gamma> \<Delta>] prems by auto
  hence "\<Gamma>;(\<Delta>\<union>\<Delta>') \<turnstile> e : \<tau>" using weakening_ty prems by auto
  thus ?thesis using prems weakening_tm by auto
qed

lemma substitution:
  fixes n :: nat
  assumes "size e = n"
  and "\<Gamma>;\<Delta> \<turnstile> e' : \<tau>'"
  and "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>"
  and "\<not> x free in e"
  shows "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e : \<tau>"
using prems proof (induct n arbitrary: e \<tau> i \<Gamma> M \<Delta> rule: nat_less_induct)
  case (1 n_y e_y \<tau>) 
  have Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" using prems by (auto simp add: ty_wf_env)
  show ?case proof (cases e_y)
    case TmUnit thus ?thesis using ty_wf_env prems by auto
  next
    case TmFreeVar thus ?thesis using ty_wf_env prems by auto
  next
    case (TmVar j)
    show "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e_y : \<tau>" proof (cases "j=i")
      case True
      hence Ht: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> TmFreeVar x : \<tau>" using prems by auto
      hence "\<tau> = \<tau>'" using wf_env_def[of "\<Gamma>,(x:\<tau>')"] by fast
      thus ?thesis using True TmVar by auto
    next
      case False
      hence "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> TmVar j : \<tau>" using prems by auto
      thus ?thesis by (cases, auto)
    qed
  next
    case (TmAbs \<tau>_y e_b)
    let ?e_bx = "vsubst_tm (TmFreeVar x) (Suc i) e_b"
    let ?e_b' = "vsubst_tm e' (Suc i) e_b"
    have Ht: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> \<lambda>[\<tau>_y]. ?e_bx : \<tau>" using prems by auto
    then obtain \<tau>_b L where Heq: "\<tau> = \<tau>_y \<rightarrow> \<tau>_b"
      and Hfin: "finite L"
      and Hwft: "\<Delta> \<Turnstile> \<tau>_y"
      and Htyp: "\<forall>z. \<not> z free in ?e_bx \<and> z \<notin> L \<and> \<not> \<Gamma>,(x:\<tau>') defines z \<longrightarrow>
                       \<Gamma>,(x:\<tau>'),(z:\<tau>_y);\<Delta> \<turnstile> freshen_tm z ?e_bx : \<tau>_b" by auto
    let ?L = "L \<union> {x} \<union> set (free_tm ?e_bx)"
    have Hfin': "finite ?L" using Hfin by auto
    have Htyp': "\<forall>z. \<not> z free in ?e_b' \<and> z \<notin> ?L \<and> \<not> \<Gamma> defines z \<longrightarrow>
                   \<Gamma>,(z:\<tau>_y);\<Delta> \<turnstile> freshen_tm z ?e_b' : \<tau>_b" proof auto
      fix z
      assume Hnzf: "\<not> z free in ?e_b'"
        and Hnzl: "z \<notin> L"
        and Hnzbx: "\<not> z free in ?e_bx"
        and Hnzx: "z \<noteq> x"
        and Hnzg: "\<not> \<Gamma> defines z"
      have A1: "size (freshen_tm z e_b) < n_y" using prems by auto
      have "\<Gamma>,(z:\<tau>_y);\<Delta> \<Turnstile> OK" using Hgok Hnzg Hwft by (auto simp add: wf_env_cons)
      hence A2: "\<Gamma>,(z:\<tau>_y);\<Delta> \<turnstile> e' : \<tau>'" using prems weakening_tm[of _ _ _ _ "{(z\<rhd>\<tau>_y)}"] by auto
      have "\<not> z free in ?e_bx" and "z \<notin> L" and "\<not> \<Gamma>,(x:\<tau>') defines z"
        using Hnzbx Hnzl Hnzg Hnzx by (auto simp add: dom_def)
      hence "\<Gamma>,(x:\<tau>'),(z:\<tau>_y);\<Delta> \<turnstile> freshen_tm z ?e_bx : \<tau>_b" using Htyp by auto
      hence A3: "\<Gamma>,(z:\<tau>_y),(x:\<tau>');\<Delta> \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z e_b) : \<tau>_b"
        by (auto simp add: env_permute subst_subst_swap[of "0" "(Suc i)"])
      have A4: "\<not> x free in (freshen_tm z e_b)" using prems by (auto simp add: not_free_subst)
      have "\<Gamma>,(z:\<tau>_y);\<Delta> \<turnstile> vsubst_tm e' (Suc i) (freshen_tm z e_b) : \<tau>_b"
        using "1.hyps" A1 A2 A3 A4 by auto
      thus  "\<Gamma>,(z:\<tau>_y);\<Delta> \<turnstile> vsubst_tm (TmFreeVar z) 0 ?e_b' : \<tau>_b"
        using prems ty_same_subst[of _ _ e'] subst_subst_swap[of "0" "Suc i"] by auto
    qed
    have "\<Gamma>;\<Delta> \<turnstile> \<lambda>[\<tau>_y]. ?e_b' : \<tau>_y \<rightarrow> \<tau>_b" using Hgok Hfin' Hwft Htyp' ..
    thus ?thesis using TmAbs Heq by auto
  next
    case (TmProd e)
    let ?e_b = "vsubst_tm e' i e"
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e"
    have "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> \<Lambda>[*]. vsubst_tm (TmFreeVar x) i e : \<tau>"
      using prems by auto
    then obtain \<tau>b L where Heq: "\<tau> = \<forall>.\<tau>b"
      and Hfin: "finite L"
      and Hval: "value ?e_bx"
      and Htyp: "\<forall> z. \<not> z free in term ?e_bx \<and> \<not> z free in type \<tau>b \<and> z \<notin> L \<and>
                 z \<notin> free_ty_env (\<Gamma>,(x:\<tau>')) \<and> \<not> \<Delta> defines z \<longrightarrow>
                   \<Gamma>,(x:\<tau>');\<Delta>,(z:\<star>) \<turnstile> (vsubst_tmty (TyFreeVar z) 0 ?e_bx) :
                                       vsubst_ty (TyFreeVar z) 0 \<tau>b" by auto
    let ?L = "L \<union> {x} \<union> set (free_ty \<tau>b) \<union> set (free_ty \<tau>') \<union> set (free_tmty ?e_bx)"
    have Hfin': "finite ?L" using Hfin by auto
    have Hval': "value (vsubst_tm e' i e)"
        using prems value_subst_value[of _ _ e] value_subst_value_rev by auto    
    have Htyp': "\<forall>\<alpha>. \<not> \<alpha> free in term ?e_b \<and> \<not> \<alpha> free in type \<tau>b \<and> \<alpha> \<notin> ?L \<and>
                      \<not> \<alpha> free in env \<Gamma> \<and> \<not> \<Delta> defines \<alpha> \<longrightarrow>
                   \<Gamma>;\<Delta>,(\<alpha>:\<star>) \<turnstile> freshen_tmty \<alpha> (vsubst_tm e' i e) : freshen_ty \<alpha> \<tau>b" proof auto
      fix \<alpha>
      assume "\<not> \<alpha> free in term ?e_b"
        and "\<alpha> \<noteq> x"
        and "\<alpha> \<notin> L"
        and "\<not> \<alpha> free in type \<tau>b"
        and "\<not> \<alpha> free in type \<tau>'"
        and "\<not> \<alpha> free in term ?e_bx"
        and "\<not> \<alpha> free in env \<Gamma>"
        and "\<not> \<Delta> defines \<alpha>"
      let ?enew = "freshen_tmty \<alpha> e"
      let ?\<tau>new = "freshen_ty \<alpha> \<tau>b"
      have A1: "size ?enew < n_y" using prems by auto
      have "\<Gamma>;\<Delta>,(\<alpha>:\<star>) \<Turnstile> OK" using ty_wf_env wf_env_ty_cons prems by auto
      hence A2: "\<Gamma>;\<Delta>,(\<alpha>:\<star>) \<turnstile> e' : \<tau>'" using weakening_ty[of \<Gamma> \<Delta> _ _ "{\<alpha>\<rhd>\<star>}"] prems by auto
      have "free_ty_env (\<Gamma>,(x:\<tau>')) = free_ty_env \<Gamma> \<union> set (free_ty \<tau>')"
        by (auto simp add: free_ty_env_def)
      hence "\<not> \<alpha> free in term ?e_bx \<and> \<not> \<alpha> free in type \<tau>b \<and> \<alpha> \<notin> L \<and>
        \<not> \<alpha> free in env (\<Gamma>,(x:\<tau>')) \<and> \<not> \<Delta> defines \<alpha>"
        using prems by auto
      hence "\<Gamma>,(x:\<tau>');\<Delta>,(\<alpha>:\<star>) \<turnstile> freshen_tmty \<alpha> (vsubst_tm (TmFreeVar x) i e) : ?\<tau>new"
        using Htyp by auto
      hence A3: "\<Gamma>,(x:\<tau>');\<Delta>,(\<alpha>:\<star>) \<turnstile> vsubst_tm (TmFreeVar x) i ?enew : ?\<tau>new"
        using subst_subst_tmty_swap[of \<alpha> "TmFreeVar x" i 0 e] by auto
      have A4: "\<not> x free in ?enew" using prems not_free_subst_sty[of x e] by auto
      have "\<Gamma>;\<Delta>,(\<alpha>:\<star>) \<turnstile> vsubst_tm e' i (freshen_tmty \<alpha> e) : freshen_ty \<alpha> \<tau>b"
        using "1.hyps" A1 A2 A3 A4 by auto
      thus "\<Gamma>;\<Delta>,(\<alpha>:\<star>) \<turnstile> vsubst_tmty (TyFreeVar \<alpha>) 0 (vsubst_tm e' i e) :
                          vsubst_ty (TyFreeVar \<alpha>) 0 \<tau>b"
        using prems by (auto simp add: ty_same_subst_ty subst_subst_tmty_swap)
    qed
    have "\<Gamma>;\<Delta> \<turnstile> \<Lambda>[*]. ?e_b : \<forall>.\<tau>b" using Hgok Hfin' Hval' Htyp' ..
    thus ?thesis using TmProd Heq by auto
  next
    case (TmApp e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmApp by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmApp by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    have "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e1x \<cdot> ?e2x : \<tau>" using prems by auto
    then obtain \<tau>1 where Ht1: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e1x : \<tau>1 \<rightarrow> \<tau>"
      and Ht2: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e2x : \<tau>1" by auto
    have "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e1 : \<tau>1 \<rightarrow> \<tau>" using "1.hyps" "1.prems" Hsize1 Hfree1 Ht1 by auto
    moreover
    have "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e2 : \<tau>1" using "1.hyps" "1.prems" Hsize2 Hfree2 Ht2 by auto
    ultimately
    show ?thesis using TmApp by auto
  next
    case (TmTapp e \<tau>a)
    have Hsize: "size e < n_y" and Hfree: "\<not> x free in e" using "1.prems" TmTapp by auto
    let ?es = "vsubst_tm (TmFreeVar x) i e"
    have "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?es \<langle>\<tau>a\<rangle> : \<tau>" using prems by auto
    then obtain \<tau>b where H1: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?es : \<forall>.\<tau>b"
      and H2: "vsubst_ty \<tau>a 0 \<tau>b = \<tau>"
      and H3: "\<Delta> \<Turnstile> \<tau>a" using typing_app_ty_elim by auto
    hence "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e : \<forall>.\<tau>b" using "1.hyps" "1.prems" Hsize Hfree by auto
    thus ?thesis using TmTapp H2 by auto
  next
    case (TmNew e_b)
    have Hsize: "size e_b < n_y" and Hfree: "\<not> x free in e_b" using "1.prems" TmNew by auto
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e_b"
    have "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> TmNew ?e_bx : \<tau>" using prems by auto
    then obtain \<tau>_b where Htb: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e_bx : \<tau>_b"
      and Heq: "\<tau> = TyRef \<tau>_b" by auto
    have "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e_b : \<tau>_b" using "1.hyps" "1.prems" Hsize Hfree Htb by auto
    thus ?thesis using TmNew Heq by auto
  next
    case (TmAssign e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmAssign by auto
    have Hfree1: "\<not> x free in e1" and Hfree2: "\<not> x free in e2" using "1.prems" TmAssign by auto
    let ?e1x = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2x = "vsubst_tm (TmFreeVar x) i e2"
    have "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e1x := ?e2x : \<tau>" using prems by auto
    then obtain \<tau>_b where Ht1: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e1x : TyRef \<tau>_b"
      and Ht2: "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e2x : \<tau>_b"
      and Heq: "\<tau> = TyUnit" by auto
    have "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e1 : TyRef \<tau>_b" using "1.hyps" "1.prems" Hsize1 Hfree1 Ht1 by auto
    moreover
    have "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e2 : \<tau>_b" using "1.hyps" "1.prems" Hsize2 Hfree2 Ht2 by auto
    ultimately
    show ?thesis using TmAssign Heq by auto
  next
    case (TmDeref e_b)
    have Hsize: "size e_b < n_y" and Hfree: "\<not> x free in e_b" using "1.prems" TmDeref by auto
    let ?e_bx = "vsubst_tm (TmFreeVar x) i e_b"
    have "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> TmDeref ?e_bx : \<tau>" using prems by auto
    hence "\<Gamma>,(x:\<tau>');\<Delta> \<turnstile> ?e_bx : TyRef \<tau>" by auto
    hence "\<Gamma>;\<Delta> \<turnstile> vsubst_tm e' i e_b : TyRef \<tau>" using "1.hyps" "1.prems" Hsize Hfree by auto
    thus ?thesis using TmDeref by auto
  next
    case (TmLoc l)
    then obtain \<tau>_b where Heq: "\<tau> = TyRef \<tau>_b" using "1.prems" by auto
    hence "\<Gamma>;\<Delta> \<turnstile> TmLoc l : TyRef \<tau>_b" using Hgok TmLoc "1.prems" by auto
    thus ?thesis using TmLoc Heq by auto
  qed
qed

lemma substitution_ty:
  fixes n :: nat
  assumes "size e = n"
  and "(vsubst_ty_env \<tau>' i \<Gamma>);\<Delta> \<Turnstile> OK"
  and "vsubst_ty_env (TyFreeVar x) i \<Gamma>;\<Delta>,(x:\<star>) \<turnstile>
          vsubst_tmty (TyFreeVar x) i e : vsubst_ty (TyFreeVar x) i \<tau>"
  and "\<not> x free in term e \<and> \<not> x free in type \<tau> \<and> x \<notin> free_ty_env \<Gamma> \<and> x \<notin> dom \<Delta>"
  and "\<Delta> \<Turnstile> \<tau>'"
  shows "vsubst_ty_env \<tau>' i \<Gamma>;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e : vsubst_ty \<tau>' i \<tau>"
using prems proof (induct n arbitrary: e \<tau> i \<tau>' \<Gamma> \<Delta> rule: nat_less_induct)
  case (1 n_y e_y \<tau>)
  let ?\<Gamma>x = "vsubst_ty_env (TyFreeVar x) i \<Gamma>"
  let ?\<Gamma>t = "vsubst_ty_env \<tau>' i \<Gamma>"
  show ?case proof (cases e_y)
    case TmUnit
    hence "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> TmUnit : vsubst_ty (TyFreeVar x) i \<tau>" using prems by auto
    hence "\<tau> = TyUnit" by (cases \<tau>, auto simp add: trivial_case)
    thus ?thesis using prems by auto
  next
    case (TmFreeVar y)
    hence Ht: "vsubst_ty_env (TyFreeVar x) i \<Gamma>;\<Delta>,(x:\<star>) \<turnstile> 
                 TmFreeVar y : vsubst_ty (TyFreeVar x) i \<tau>"
      using prems by auto
    hence Hx': "?\<Gamma>x = vsubst_ty_env (TyFreeVar x) i (\<Gamma>,(y:\<tau>))"
      by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> x free in env (\<Gamma>,(y:\<tau>))" using prems by (auto simp add: free_ty_env_def)
    moreover
    hence "\<not> x free in env \<Gamma>" using prems by auto
    ultimately
    have "?\<Gamma>t = vsubst_ty_env \<tau>' i \<Gamma>,(y:vsubst_ty \<tau>' i \<tau>)"
      using subst_subst_equal_env[of x i \<Gamma> "\<Gamma>,(y:\<tau>)"] by (auto simp add: vsubst_ty_env_def)
    thus ?thesis using prems by auto
  next
    case (TmVar j)
    hence "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> vsubst_tmty (TyFreeVar x) i (TmVar j) : vsubst_ty (TyFreeVar x) i \<tau>"
      using prems by auto
    hence "?\<Gamma>t;\<Delta>,(x:\<star>) \<turnstile> TmVar j : vsubst_ty (TyFreeVar x) i \<tau>" by auto
    thus ?thesis by auto
  next
    case (TmApp e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmApp by auto
    let ?e1s = "vsubst_tmty (TyFreeVar x) i e1"
    let ?e2s = "vsubst_tmty (TyFreeVar x) i e2"
    have "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e1s \<cdot> ?e2s : vsubst_ty (TyFreeVar x) i \<tau>"
      using prems by auto
    then obtain \<tau>new where H1: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e1s : \<tau>new \<rightarrow> vsubst_ty (TyFreeVar x) i \<tau>"
      and H2: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e2s : \<tau>new" by auto
    hence "\<Delta>,(x:\<star>) \<Turnstile> \<tau>new" using ty_wf_type by auto
    then obtain \<tau>n where Hn: "\<tau>new = vsubst_ty (TyFreeVar x) i \<tau>n"
      and Hfree: "\<not> x free in type \<tau>n" using parametrize_ty[of x _ \<tau>new i] by auto
    have "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e1s : vsubst_ty (TyFreeVar x) i (\<tau>n \<rightarrow> \<tau>)"
      using H1 Hn by auto
    moreover
    have "\<not> x free in term e1 \<and> x \<notin> set (free_ty (\<tau>n \<rightarrow> \<tau>)) \<and> x \<notin> free_ty_env \<Gamma>
            \<and> x \<notin> Environ.dom \<Delta>" using prems Hfree by auto
    ultimately
    have Ht1: "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e1 : vsubst_ty \<tau>' i (\<tau>n \<rightarrow> \<tau>)"
        using H1 Hsize1 "1.hyps" "1.prems" by blast
    moreover
    have "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e2s : vsubst_ty (TyFreeVar x) i \<tau>n " using H2 Hn by auto
    moreover
    have "\<not> x free in term e2 \<and> x \<notin> set (free_ty \<tau>n) \<and> x \<notin> free_ty_env \<Gamma>
            \<and> x \<notin> Environ.dom \<Delta>" using prems Hfree by auto
    ultimately
    have Ht2: "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e2 : vsubst_ty \<tau>' i \<tau>n"
      using H2 Hsize2 "1.hyps" "1.prems" by blast
    show ?thesis using TmApp Ht1 Ht2 by auto
  next
    case (TmTapp e \<tau>a)
    let ?es = "vsubst_tmty (TyFreeVar x) i e"
    let ?\<tau>s = "vsubst_ty (TyFreeVar x) i \<tau>a"
    have Ht1: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?es\<langle>?\<tau>s\<rangle> : vsubst_ty (TyFreeVar x) i \<tau>" using prems by auto
    then obtain \<tau>b where H1: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?es : \<forall>.\<tau>b"
      and H2: "vsubst_ty (TyFreeVar x) i \<tau> = vsubst_ty ?\<tau>s 0 \<tau>b"
      and H3: "\<Delta>,(x:\<star>) \<Turnstile> ?\<tau>s" by auto
    have "\<Delta>,(x:\<star>) \<Turnstile> \<forall>.\<tau>b" using H1 ty_wf_type by auto
    then obtain \<tau>n' where Hn: "vsubst_ty (TyFreeVar x) i \<tau>n' = \<forall>.\<tau>b"
      and  Hfree: "\<not> x free in type \<tau>n'" using parametrize_ty[of x \<Delta> "\<forall>.\<tau>b" i] by auto
    then obtain \<tau>n where Hn2: "\<tau>n' = \<forall>.\<tau>n" by (cases \<tau>n', auto simp add: trivial_case)
    hence H4: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?es : vsubst_ty (TyFreeVar x) i \<tau>n'"
      and H5: "\<forall>.\<tau>b = \<forall>.vsubst_ty (TyFreeVar x) (Suc i) \<tau>n" using H1 Hn by auto
    have Hsize: "size e < n_y" using "1.prems" TmTapp by auto
    have "\<not> x free in term e \<and> x \<notin> set (free_ty \<tau>n') \<and> x \<notin> free_ty_env \<Gamma>
      \<and> x \<notin> Environ.dom \<Delta>" using prems Hfree by auto     
    hence "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e : vsubst_ty \<tau>' i \<tau>n'"
      using H4 Hn "1.hyps" "1.prems" Hsize by auto
    hence Htfinal: "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e : \<forall>.vsubst_ty \<tau>' (Suc i) \<tau>n" using Hn2 by auto
    hence Heqt: "vsubst_ty (TyFreeVar x) i \<tau> =
        vsubst_ty (vsubst_ty (TyFreeVar x) i \<tau>a) 0 (vsubst_ty (TyFreeVar x) (Suc i) \<tau>n)"
      using H2 H5  H3 subst_var_type_ok[of x \<Delta> i \<tau>a] prems by auto
    have "\<Delta>,(x:\<star>) \<Turnstile> vsubst_ty (TyFreeVar x) i \<tau>" using Ht1 ty_wf_type by auto
    hence "set (free_db_ty (vsubst_ty (TyFreeVar x) i \<tau>)) = {}"
      using ok_no_free_db by auto
    hence A1: "set (free_db_ty \<tau>) \<subseteq> {i}" using vsubst_ty_free_db_ty by auto
    have "set (free_db_ty (vsubst_ty (TyFreeVar x) i \<tau>a)) = {}"
      using ok_no_free_db H3 by auto
    moreover
    have "set (free_db_ty (vsubst_ty (TyFreeVar x) i \<tau>a)) =
      set (free_db_ty \<tau>a) - {i}"
      using vsubst_ty_free_db_ty by auto
    ultimately
    have A2: "set (free_db_ty \<tau>a) \<subseteq> {i}" by auto
    have "\<Delta>,(x:\<star>) \<Turnstile>
      vsubst_ty (vsubst_ty (TyFreeVar x) i \<tau>a) 0 (vsubst_ty (TyFreeVar x) (Suc i) \<tau>n)"
      using Ht1 Heqt ty_wf_type by auto
    hence "set (free_db_ty
      (vsubst_ty (vsubst_ty (TyFreeVar x) i \<tau>a) 0 (vsubst_ty (TyFreeVar x) (Suc i) \<tau>n))) = {}"
      using ok_no_free_db by auto
    moreover
    have "set (free_db_ty
      (vsubst_ty (vsubst_ty (TyFreeVar x) i \<tau>a) 0 (vsubst_ty (TyFreeVar x) (Suc i) \<tau>n)))
      = set (free_db_ty ((vsubst_ty (TyFreeVar x) (Suc i) \<tau>n))) - {0}"
      using vsubst_ty_free_db_ty[of "vsubst_ty (TyFreeVar x) i \<tau>a"]
        ok_no_free_db[of _ "vsubst_ty (TyFreeVar x) i \<tau>a"] prems H3 by auto 
    ultimately
    have A3: "set (free_db_ty \<tau>n) \<subseteq> {0, Suc i}" using vsubst_ty_free_db_ty by auto
    have Hwft: "\<Delta> \<Turnstile> vsubst_ty \<tau>' i \<tau>a" using H3 subst_var_type_ok[of x \<Delta> i \<tau>a] prems by auto
    have "set (free_db_ty \<tau>') = {}" and "x \<notin> set (free_ty \<tau>)" and "x \<notin> set (free_ty \<tau>a)"
      and  "x \<notin> set (free_ty \<tau>n)" using prems ok_no_free_db by auto
    hence "vsubst_ty \<tau>' i \<tau> = vsubst_ty (vsubst_ty \<tau>' i \<tau>a) 0 (vsubst_ty \<tau>' (Suc i) \<tau>n)"
      using Heqt A1 A2 A3 vsubst_equal_sanity[of x i \<tau>a \<tau>n \<tau> \<tau>'] by auto
    thus "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e_y : vsubst_ty \<tau>' i \<tau>"
      using TmTapp Htfinal Hwft by auto
  next
    case (TmNew er)
    have Hsize: "size er < n_y" using "1.prems" TmNew by auto
    let ?es = "vsubst_tmty (TyFreeVar x) i er"
    have "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> TmNew ?es : vsubst_ty (TyFreeVar x) i \<tau>" using prems by auto
    then obtain \<tau>r where H1: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?es : vsubst_ty (TyFreeVar x) i \<tau>r" and
      H2: "TyRef \<tau>r = \<tau>" by (cases \<tau>, auto simp add: trivial_case)
    have "\<not> x free in term er \<and> \<not> x free in type \<tau>r" using prems by (auto simp add: mem_iff)
    moreover
    have "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?es : vsubst_ty (TyFreeVar x) i \<tau>r" using H1 by auto
    ultimately
    have "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i er : vsubst_ty \<tau>' i \<tau>r"
      using "1.hyps" "1.prems" Hsize by auto
    thus ?thesis using TmNew H2 by auto
  next
    case (TmAssign e1 e2)
    have Hsize1: "size e1 < n_y" and Hsize2: "size e2 < n_y" using "1.prems" TmAssign by auto
    let ?e1s = "vsubst_tmty (TyFreeVar x) i e1"
    let ?e2s = "vsubst_tmty (TyFreeVar x) i e2"
    have Ht: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e1s := ?e2s : vsubst_ty (TyFreeVar x) i \<tau>" using prems by auto
    hence H3: "\<tau> = TyUnit" by (cases \<tau>, auto simp add: trivial_case)
    from Ht obtain \<tau>r' where H1: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e1s : TyRef \<tau>r'"
      and H2: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e2s : \<tau>r'" by auto
    hence "\<Delta>,(x:\<star>) \<Turnstile> \<tau>r'" using ty_wf_type by auto
    then obtain \<tau>r where "\<tau>r' = vsubst_ty (TyFreeVar x) i \<tau>r" and "\<not> x free in type \<tau>r"
      using parametrize_ty[of x \<Delta> \<tau>r'] by auto
    hence H1': "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e1s : vsubst_ty (TyFreeVar x) i (TyRef \<tau>r)"
      and H2': "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?e2s : vsubst_ty (TyFreeVar x) i \<tau>r"
      using H1 H2 by auto    
    have "\<not> x free in term e1 \<and> \<not> x free in type (TyRef \<tau>r)"
      using prems by (auto simp add: mem_iff)
    hence Ht1: "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e1 : vsubst_ty \<tau>' i (TyRef \<tau>r)"
      using H1' Hsize1 "1.hyps" "1.prems" by blast
    have "\<not> x free in term e2 \<and> \<not> x free in type \<tau>r"
      using prems by (auto simp add: mem_iff)
    hence Ht2: "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i e2 : vsubst_ty \<tau>' i \<tau>r"
      using H2' Hsize2 "1.hyps" "1.prems" by auto
    show ?thesis using TmAssign H3 Ht1 Ht2 by auto
  next
    case (TmDeref er)
    have Hsize: "size er < n_y" using "1.prems" TmDeref by auto
    let ?es = "vsubst_tmty (TyFreeVar x) i er"
    have "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> TmDeref ?es : vsubst_ty (TyFreeVar x) i \<tau>" using prems by auto
    then obtain \<tau>r' where H1: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?es : TyRef \<tau>r'" and
      H2: "\<tau>r' = vsubst_ty (TyFreeVar x) i \<tau>" by auto
    hence "\<Delta>,(x:\<star>) \<Turnstile> \<tau>r'" using ty_wf_type by auto
    then obtain \<tau>r where Hr: "\<tau>r' = vsubst_ty (TyFreeVar x) i \<tau>r" and Hfr: "\<not> x free in type \<tau>r"
      using parametrize_ty[of x \<Delta> \<tau>r'] by auto
    hence Heq: "\<tau>r = \<tau>" using H2 Hfr subst_subst_equal_tyty[of x i \<tau>r \<tau>] prems by blast
    have H1': "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> ?es : vsubst_ty (TyFreeVar x) i (TyRef \<tau>r)"
      using H1 Hr by auto
    have "\<not> x free in term er \<and> \<not> x free in type (TyRef \<tau>r)"
      using prems by (auto simp add: mem_iff)
    hence "?\<Gamma>t;\<Delta> \<turnstile> vsubst_tmty \<tau>' i er : vsubst_ty \<tau>' i (TyRef \<tau>r)"
      using H1' Hsize "1.hyps" "1.prems" by blast
    thus ?thesis using TmDeref Heq H2 by auto
  next
    case (TmLoc l)
    hence Ht: "vsubst_ty_env (TyFreeVar x) i \<Gamma>;\<Delta>,(x:\<star>) \<turnstile> TmLoc l : vsubst_ty (TyFreeVar x) i \<tau>"
      using prems by auto
    then obtain \<tau>ins where Heq: "\<tau> = TyRef \<tau>ins" by (cases \<tau>, auto simp add: trivial_case)
    hence "(l\<rhd>vsubst_ty (TyFreeVar x) i \<tau>ins) \<in> ?\<Gamma>x" using Ht by auto
    hence "?\<Gamma>x = ?\<Gamma>x,(l:vsubst_ty (TyFreeVar x) i \<tau>ins)" by auto
    hence Hx': "?\<Gamma>x = vsubst_ty_env (TyFreeVar x) i (\<Gamma>,(l:\<tau>ins))"
      by (auto simp add: vsubst_ty_env_def)
    moreover
    have "free_ty_env (\<Gamma>,(l:\<tau>ins)) = free_ty_env \<Gamma> \<union> set (free_ty \<tau>ins)"
      using free_ty_env_def by auto
    hence Hx'': "\<not> x free in env \<Gamma>" and Hx''': "\<not> x free in env (\<Gamma>,(l:\<tau>ins))"
      using prems by auto
    ultimately
    have "?\<Gamma>t = vsubst_ty_env \<tau>' i (\<Gamma>,(l:\<tau>ins))"
      using subst_subst_equal_env[of x i \<Gamma> "\<Gamma>,(l:\<tau>ins)"] by auto
    hence "?\<Gamma>t = vsubst_ty_env \<tau>' i \<Gamma>,(l:vsubst_ty \<tau>' i \<tau>ins)"
      by (auto simp add: vsubst_ty_env_def)
    hence "(l\<rhd>vsubst_ty \<tau>' i \<tau>ins) \<in> ?\<Gamma>t" by auto
    thus ?thesis using prems by auto
  next
    case (TmProd e)
    have Ht: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> \<Lambda>[*]. vsubst_tmty (TyFreeVar x) (Suc i) e :
                            vsubst_ty (TyFreeVar x) i \<tau>"
      using prems by auto
    then obtain \<tau>b where Heqb: "\<tau> = \<forall>.\<tau>b" by (cases \<tau>, auto simp add: trivial_case)
    hence Ht: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> \<Lambda>[*]. vsubst_tmty (TyFreeVar x) (Suc i) e :
                             \<forall>. vsubst_ty (TyFreeVar x) (Suc i) \<tau>b" using Ht by auto
    thus ?thesis proof
      fix L \<tau>r
      assume "?\<Gamma>x;\<Delta>,(x:\<star>) \<Turnstile> OK"
        and "finite L"
        and Hv: "value (vsubst_tmty (TyFreeVar x) (Suc i) e)"
        and Ht: "\<forall>a. a \<notin> set (free_tmty (vsubst_tmty (TyFreeVar x) (Suc i) e)) \<and>
                     a \<notin> set (free_ty \<tau>r) \<and> a \<notin> L \<and> a \<notin> free_ty_env ?\<Gamma>x \<and> a \<notin> dom (\<Delta>,(x:\<star>)) \<longrightarrow>
                      ?\<Gamma>x;\<Delta>,(x:\<star>),(a:\<star>) \<turnstile> vsubst_tmty (TyFreeVar a) 0
                      (vsubst_tmty (TyFreeVar x) (Suc i) e) : vsubst_ty (TyFreeVar a) 0 \<tau>r"
        and "\<forall>.vsubst_ty (TyFreeVar x) (Suc i) \<tau>b = \<forall>.\<tau>r"
      hence Heq: "\<tau>r = vsubst_ty (TyFreeVar x) (Suc i) \<tau>b" by auto
      have "?\<Gamma>t;\<Delta> \<Turnstile> OK" using prems by auto
      moreover
      have "finite ({x} \<union> set (free_ty \<tau>r) \<union> L \<union>
              set (free_tmty (vsubst_tmty (TyFreeVar x) (Suc i) e)))"
        by auto
      moreover
      have "value (vsubst_tmty \<tau>' (Suc i) e)" using Hv by (induct e arbitrary: i, auto)
      ultimately
      have "?\<Gamma>t;\<Delta> \<turnstile> \<Lambda>[*]. vsubst_tmty \<tau>' (Suc i) e : \<forall>.(vsubst_ty \<tau>' (Suc i) \<tau>b)" proof (rule, auto)
        fix a
        assume "a \<notin> set (free_tmty (vsubst_tmty \<tau>' (Suc i) e))"
          and Hfr1: "a \<notin> set (free_ty (vsubst_ty \<tau>' (Suc i) \<tau>b))"
          and Hneq: "a \<noteq> x" and Hfr2: "a \<notin> set (free_ty \<tau>r)" and Hnl: "a \<notin> L"
          and Hfr3: "a \<notin> set (free_tmty (vsubst_tmty (TyFreeVar x) (Suc i) e))"
          and Hfr4: "a \<notin> free_ty_env ?\<Gamma>t"
          and Hfr5: "a \<notin> Environ.dom \<Delta>"
        let ?\<Gamma>i = "convert_ty_env (Suc i) i \<Gamma>"
        have "\<Delta>,(x:\<star>) \<Turnstile> TyFreeVar x" using prems wf_env_wf_env_ty wf_tvar by auto
        hence Hgisx: "?\<Gamma>x = vsubst_ty_env (TyFreeVar x) (Suc i) ?\<Gamma>i"
          using subst_convert_env[of _ "\<Delta>,(x:\<star>)" x i \<Gamma> "TyFreeVar x" "Suc i"] prems by auto
        have Hgist: "?\<Gamma>t = vsubst_ty_env \<tau>' (Suc i) ?\<Gamma>i"
          using wf_tenv_cons[of \<Delta> _ x] subst_convert_env[of _ "\<Delta>,(x:\<star>)" x i \<Gamma> \<tau>' "Suc i"] prems
          by auto
        obtain n where A1: "size (vsubst_tmty (TyFreeVar x) (Suc i) e) = n" by auto
        hence A2: "n < n_y" using prems by auto
        have A3:  "\<not> x free in term (freshen_tmty a e)" using not_free_subst_tmty prems by auto
        have A4: "\<not> x free in type (freshen_ty a \<tau>b)" using not_free_subst_ty prems by auto
        have A5: "\<not> \<Delta>,(a:\<star>) defines x" using prems by (auto simp add: dom_def)
        have A6:  "x \<notin> free_ty_env ?\<Gamma>i" using prems not_free_convert_env by auto
        have A7:  "\<Delta>,(a:\<star>) \<Turnstile> \<tau>'" using prems wf_tenv_cons by auto
        have A8:  "vsubst_ty_env \<tau>' (Suc i) ?\<Gamma>i;\<Delta>,(a:\<star>) \<Turnstile> OK"
          using Hgist wf_env_ty_cons prems by auto
        have "\<not> a free in env ?\<Gamma>x" using not_free_subst_env[of a \<Gamma> "TyFreeVar x" i]
          inv_not_free_subst_env[of a \<tau>' i \<Gamma>] prems by auto
        hence "?\<Gamma>x;\<Delta>,(x:\<star>),(a:\<star>) \<turnstile> freshen_tmty a (vsubst_tmty (TyFreeVar x) (Suc i) e) :
          freshen_ty a (vsubst_ty (TyFreeVar x) (Suc i) \<tau>b)"
          using Ht Hfr1 Hfr2 Hfr3 Hfr4 Hfr5 Hnl Heq Hneq by (auto simp add: dom_def)
        hence A9: "vsubst_ty_env (TyFreeVar x) (Suc i) ?\<Gamma>i;\<Delta>,(a:\<star>),(x:\<star>) \<turnstile>
                     vsubst_tmty (TyFreeVar x) (Suc i) (freshen_tmty a e) :
                     vsubst_ty (TyFreeVar x) (Suc i) (freshen_ty a \<tau>b)"
          using subst_subst_ty_swap[of 0 "Suc i"] subst_subst_tyty_swap[of 0 "Suc i"] 
            Hgisx env_permute[of x _ a _ \<Delta>] by auto
        have "vsubst_ty_env \<tau>' (Suc i) ?\<Gamma>i;\<Delta>,(a:\<star>) \<turnstile> vsubst_tmty \<tau>' (Suc i)
          (freshen_tmty a e) : vsubst_ty \<tau>' (Suc i) (freshen_ty a \<tau>b)"
          using A1 A2 A3 A4 A5 A6 A7 A8 A9 "1.hyps" by auto
        thus "vsubst_ty_env \<tau>' i \<Gamma>;\<Delta>,(a:\<star>) \<turnstile>
          vsubst_tmty (TyFreeVar a) 0 (vsubst_tmty \<tau>' (Suc i) e) :
          vsubst_ty (TyFreeVar a) 0 (vsubst_ty \<tau>' (Suc i) \<tau>b)"
          using prems subst_subst_ty_swap[of 0 "Suc i"] subst_subst_tyty_swap[of "Suc i" 0 a \<tau>' e]
            no_free_db_vars_same_subst_ty ok_no_free_db Hgist by auto
      qed
      thus ?thesis using TmProd Heqb by auto
    qed
  next
    case (TmAbs \<tau>a e)
    have Ht: "?\<Gamma>x;\<Delta>,(x:\<star>) \<turnstile> \<lambda>[vsubst_ty (TyFreeVar x) i \<tau>a]. vsubst_tmty (TyFreeVar x) i e
                          : vsubst_ty (TyFreeVar x) i \<tau>"
      using prems by auto
    thus ?thesis proof
      fix L \<tau>r'
      assume "?\<Gamma>x;\<Delta>,(x:\<star>) \<Turnstile> OK"
        and "finite L"
        and "\<Delta>,(x:\<star>) \<Turnstile> vsubst_ty (TyFreeVar x) i \<tau>a"
        and Ht: "\<forall>z. z \<notin> set (free_tm (vsubst_tmty (TyFreeVar x) i e)) \<and>
                 z \<notin> L \<and> z \<notin> Environ.dom ?\<Gamma>x \<longrightarrow>
                 ?\<Gamma>x,(z:vsubst_ty (TyFreeVar x) i \<tau>a);\<Delta>,(x:\<star>) \<turnstile> vsubst_tm (TmFreeVar z) 0
                                         (vsubst_tmty (TyFreeVar x) i e) : \<tau>r'"
        and Heq: "vsubst_ty (TyFreeVar x) i \<tau> = vsubst_ty (TyFreeVar x) i \<tau>a \<rightarrow> \<tau>r'"
      hence "\<Delta>,(x:\<star>) \<Turnstile> vsubst_ty (TyFreeVar x) i \<tau>a \<rightarrow> \<tau>r'" using prems ty_wf_type by auto
      hence "\<Delta>,(x:\<star>) \<Turnstile> \<tau>r'" by auto
      then obtain \<tau>r where Htr: "\<tau>r' = vsubst_ty (TyFreeVar x) i \<tau>r" and Hfx: "\<not> x free in type \<tau>r"
        using parametrize_ty[of _ \<Delta> \<tau>r'] by blast
      hence "vsubst_ty (TyFreeVar x) i \<tau> = vsubst_ty (TyFreeVar x) i (\<tau>a \<rightarrow> \<tau>r)" using Heq by auto
      moreover
      hence "\<not> x free in type (\<tau>a \<rightarrow> \<tau>r)" using prems Hfx by auto
      ultimately
      have Hteq: "\<tau> = \<tau>a \<rightarrow> \<tau>r" using Heq subst_subst_equal_tyty[of x i \<tau> "\<tau>a\<rightarrow>\<tau>r"] prems by auto
      let ?L' = "set (free_tm (vsubst_tmty (TyFreeVar x) i e)) \<union> L \<union> dom ?\<Gamma>x"
      have "?\<Gamma>t;\<Delta> \<Turnstile> OK" by assumption
      moreover
      have "finite ?L'" using prems wf_env_fin by auto
      moreover
      have "\<Delta> \<Turnstile> vsubst_ty \<tau>' i \<tau>a" using subst_var_type_ok[of x \<Delta> i \<tau>a \<tau>'] prems by auto
      ultimately
      have "?\<Gamma>t;\<Delta> \<turnstile> \<lambda>[vsubst_ty \<tau>' i \<tau>a]. vsubst_tmty \<tau>' i e :
        (vsubst_ty \<tau>' i \<tau>a) \<rightarrow> (vsubst_ty \<tau>' i \<tau>r)" proof (rule, auto)
        fix z
        assume Hz: "z \<notin> set (free_tm (vsubst_tmty \<tau>' i e))"
          and Hz1: "z \<notin> L" and Hz2: "z \<notin> dom ?\<Gamma>x" and Hz3: "z \<notin> dom ?\<Gamma>t"
          and Hz4: "z \<notin> set (free_tm (vsubst_tmty (TyFreeVar x) i e))"
        obtain n where A1: "size (vsubst_tm (TmFreeVar z) 0 e) = n" by auto
        hence A2: "n < n_y" using prems by auto
        have A3:  "\<not> x free in term (freshen_tm z e)" using not_free_subst_tm prems by auto
        have A4: "\<not> x free in type \<tau>r" using prems by auto
        have A5: "\<Delta> \<Turnstile> \<tau>'" by assumption
        have "free_ty_env (\<Gamma>,(z:\<tau>a)) = free_ty_env \<Gamma> \<union> set (free_ty \<tau>a)"
          by (auto simp add: free_ty_env_def)
        hence A6: "x \<notin> free_ty_env (\<Gamma>,(z:\<tau>a))" using prems by (auto simp add: free_ty_env_def)
        have A7: "\<not> \<Delta> defines x" using prems by auto
        have "?\<Gamma>x,(z:vsubst_ty (TyFreeVar x) i \<tau>a);\<Delta>,(x:\<star>) \<turnstile>
          freshen_tm z (vsubst_tmty (TyFreeVar x) i e) :
          vsubst_ty (TyFreeVar x) i \<tau>r"
          using Hz1 Hz2 Hz4 Htr Ht by auto
        have "?\<Gamma>x,(z:vsubst_ty (TyFreeVar x) i \<tau>a);\<Delta>,(x:\<star>) \<turnstile>
          vsubst_tmty (TyFreeVar x) i (freshen_tm z e) :
          vsubst_ty (TyFreeVar x) i \<tau>r" 
          using Hz1 Hz2 Hz4 Htr Ht subst_subst_tmty_swap[of _ _ 0 i] by auto
        also
        have "vsubst_ty_env (TyFreeVar x) i \<Gamma>,(z:vsubst_ty (TyFreeVar x) i \<tau>a) =
             vsubst_ty_env (TyFreeVar x) i (\<Gamma>,(z:\<tau>a))"
          by (auto simp add: vsubst_ty_env_def)
        finally
        have A8: "vsubst_ty_env (TyFreeVar x) i (\<Gamma>,(z:\<tau>a));\<Delta>,(x:\<star>) \<turnstile>
          vsubst_tmty (TyFreeVar x) i (freshen_tm z e) : vsubst_ty (TyFreeVar x) i \<tau>r"
          by auto
        have H:"vsubst_ty_env \<tau>' i (\<Gamma>,(z:\<tau>a)) = (vsubst_ty_env \<tau>' i \<Gamma>),(z:vsubst_ty \<tau>' i \<tau>a)"
          using vsubst_ty_env_def by auto
        moreover
        have "?\<Gamma>t;\<Delta> \<Turnstile> OK" using prems ty_wf_env by auto
        moreover
        have "\<Delta> \<Turnstile> vsubst_ty \<tau>' i \<tau>a" using subst_var_type_ok[of x \<Delta> i] prems by auto
        moreover
        have "\<not> ?\<Gamma>t defines z" using prems by auto
        ultimately
        have A9: "vsubst_ty_env \<tau>' i (\<Gamma>,(z:\<tau>a));\<Delta> \<Turnstile> OK" using wf_env_cons by auto
        have "vsubst_ty_env \<tau>' i (\<Gamma>,(z:\<tau>a));\<Delta> \<turnstile>
              vsubst_tmty \<tau>' i (freshen_tm z e) : vsubst_ty \<tau>' i \<tau>r"
          using "1.hyps" A1 A2 A3 A4 A5 A6 A7 A8 A9 by auto
        also
        have "vsubst_ty_env \<tau>' i (\<Gamma>,(z:\<tau>a)) = vsubst_ty_env \<tau>' i \<Gamma>,(z:vsubst_ty \<tau>' i \<tau>a)"
          by (auto simp add: vsubst_ty_env_def)
        finally
        have "?\<Gamma>t,(z:vsubst_ty \<tau>' i \<tau>a);\<Delta> \<turnstile> freshen_tm z (vsubst_tmty \<tau>' i e) : vsubst_ty \<tau>' i \<tau>r"
          using subst_subst_tytm_swap by auto
        thus "?\<Gamma>t,(z:vsubst_ty \<tau>' i \<tau>a);\<Delta> \<turnstile>
        vsubst_tm (TmFreeVar z) 0 (vsubst_tmty \<tau>' i e) : vsubst_ty \<tau>' i \<tau>r" by auto
      qed
      thus ?thesis using prems Hteq by auto
    qed
  qed
qed

constdefs
  storeTyping :: "store \<Rightarrow> tm_env \<Rightarrow> ty_env \<Rightarrow> bool"   ("\<Turnstile> _:_;_ " [50,50] 50)
  "storeTyping S \<Gamma> \<Delta> \<equiv> S \<Turnstile> Store \<and> (\<forall>x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<longrightarrow> (\<exists>v. (x\<mapsto>v) \<in> S \<and> \<Gamma>;\<Delta> \<turnstile> v : \<tau>))"

lemma store_ty_store:
  assumes "\<Turnstile> S : \<Gamma>;\<Delta>"
  shows "S \<Turnstile> Store"
using prems by (auto simp add: storeTyping_def)

lemma store_ty_finite:
  assumes "\<Turnstile> S : \<Gamma>;\<Delta>"
  shows "finite S"
using prems by (auto simp add: storeTyping_def wf_store_def)

lemma st_ty_update:
  assumes "\<Turnstile> S : \<Gamma>;\<Delta>"
  and "\<Gamma>;\<Delta> \<turnstile> TmFreeVar z : \<tau>"
  and "\<Gamma>;\<Delta> \<turnstile> v : \<tau>"
  and "value v"
  shows "\<Turnstile> update S z v : \<Gamma>;\<Delta>"
proof (auto simp add: storeTyping_def)
  fix xa \<tau>a
  assume "xa\<rhd>\<tau>a \<in> \<Gamma>"
  show "\<exists>va. (xa\<mapsto>va) \<in> update S z v \<and> \<Gamma>;\<Delta> \<turnstile> va : \<tau>a" proof (cases "z = xa")
    case True
    hence "(xa\<mapsto>v) \<in> update S z v" using prems update_def by auto
    moreover
    have "\<tau> = \<tau>a" using prems wf_env_def[of \<Gamma>] by blast
    ultimately
    show ?thesis using prems by auto
  next
    case False show ?thesis using prems by (auto simp add: update_def storeTyping_def)
  qed
next
  show "update S z v \<Turnstile> Store" using prems store_ty_store store_update_ok by auto
qed

lemma store_dom_in:
  assumes "\<Turnstile> S : \<Gamma>;\<Delta>"
  and "\<Gamma> defines z"
  shows "S defines z"
using prems proof -
  obtain b where "(z,b) \<in> \<Gamma>" using prems by (auto simp add: dom_def)
  show ?thesis proof (cases b)
    fix \<tau>
    assume "b = Term \<tau>"
    from storeTyping_def obtain v where "z\<mapsto>v \<in> S" using prems by blast
    thus "S defines z" using prems by (auto simp add: dom_def)
  qed
qed

lemma store_dom_not_in:
  assumes "\<Turnstile> S : \<Gamma>;\<Delta>"
  and "\<not> S defines z"
  shows "\<not> \<Gamma> defines z"
using prems by (auto simp add: store_dom_in)

lemma store_typing_elim:
  assumes "\<Turnstile> S : \<Gamma>;\<Delta>"
  and "(z\<mapsto>v) \<in> S"
  and "(z\<rhd>\<tau>) \<in> \<Gamma>"
  shows "\<Gamma>;\<Delta> \<turnstile> v : \<tau>"
using prems storeTyping_def[of S \<Gamma>] wf_store_def[of S] by fast

lemma store_typing_cons:
  assumes "\<Turnstile> S : \<Gamma>;\<Delta>"
  and "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  and "value e"
  and "\<not> S defines z"
  shows "\<Turnstile> S,(z\<mapsto>e) : \<Gamma>,(z:\<tau>);\<Delta>"
proof (unfold storeTyping_def, rule)
  show "S,(z\<mapsto>e) \<Turnstile> Store" using prems by (auto simp add: storeTyping_def wf_store_cons)
next
  have Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" using prems ty_wf_env by auto
  have Hgnz: "\<not> \<Gamma> defines z" using prems store_dom_not_in by auto
  have "\<Delta> \<Turnstile> \<tau>" using prems ty_wf_type by auto
  hence Hgokz: "\<Gamma>,(z:\<tau>);\<Delta> \<Turnstile> OK" using Hgok Hgnz by (auto simp add: wf_env_cons)
  { fix xa \<tau>a
    assume H1: "(xa\<rhd>\<tau>a) \<in> \<Gamma>,(z:\<tau>)"
    have "\<exists>v. (xa\<mapsto>v) \<in> S,(z\<mapsto>e) \<and> \<Gamma>,(z:\<tau>);\<Delta> \<turnstile> v : \<tau>a" proof (cases "z=xa")
      case True
      hence "\<tau> = \<tau>a" using Hgokz H1 wf_env_def[of "\<Gamma>,(z:\<tau>)"] by blast
      hence "\<Gamma>;\<Delta> \<turnstile> e : \<tau>a" using prems by auto
      hence "\<Gamma>,(z:\<tau>);\<Delta> \<turnstile> e : \<tau>a" using prems weakening_tm[of _ _ _ _ "{z\<rhd>\<tau>}"] Hgokz by auto
      thus ?thesis using True by auto
    next
      case False
      hence "(xa\<rhd>\<tau>a) \<in> \<Gamma>" using H1 by auto
      then obtain v where "(xa\<mapsto>v) \<in> S \<and> \<Gamma>;\<Delta> \<turnstile> v : \<tau>a" using storeTyping_def[of S \<Gamma>] prems by blast
      hence "(xa\<mapsto>v) \<in> S,(z\<mapsto>e) \<and> \<Gamma>,(z:\<tau>);\<Delta> \<turnstile> v : \<tau>a" using weakening_tm[of _ _ _ _ "{z\<rhd>\<tau>}"] Hgokz by auto
      thus ?thesis by auto
    qed
  }
  thus "\<forall>x \<tau>'. x\<rhd>\<tau>' \<in> \<Gamma>,(z:\<tau>) \<longrightarrow> (\<exists>v. (x\<mapsto>v) \<in> S,(z\<mapsto>e) \<and> \<Gamma>,(z:\<tau>);\<Delta> \<turnstile> v : \<tau>')" by auto
qed

theorem preservation:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  and "S;e \<hookrightarrow> S';e'"
  and "\<Turnstile> S : \<Gamma>;\<Delta>"
  shows "\<exists> \<Gamma>'. \<Turnstile> S' : (\<Gamma>\<union>\<Gamma>');\<Delta> \<and> (\<Gamma>\<union>\<Gamma>');\<Delta> \<turnstile> e' : \<tau>"
proof -
  have "\<Gamma>;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile> \<tau>" and "\<Delta> \<Turnstile> TY"
    using prems ty_wf_env ty_wf_type ty_wf_env_ty by auto
  from prems this show ?thesis proof (induct arbitrary: e')
  case (t_unit \<Delta> \<Gamma> e')
  have "S;TmUnit \<hookrightarrow> S';e'" .
  thus ?case proof cases
    case (e_val Sa v z)
    hence "\<Gamma>,(z:TyUnit);\<Delta> \<Turnstile> OK" using t_unit by (auto simp add: wf_env_cons store_dom_not_in)
    hence "\<Gamma>,(z:TyUnit);\<Delta> \<turnstile> TmFreeVar z : TyUnit" using t_unit by auto
    moreover
    have "\<Turnstile> S,(z\<mapsto>TmUnit) : \<Gamma>,(z:TyUnit); \<Delta>"
      using t_unit e_val store_typing_cons[of _ _ _ TmUnit TyUnit] by auto
    ultimately
    show ?thesis using e_val by (auto simp add: exI[of _ "{(z\<rhd>TyUnit)}"])
  qed auto
next
  case (t_var \<Delta> \<Gamma> \<tau> x e')
  have "S;TmFreeVar x \<hookrightarrow> S';e'" .
  thus ?case by (cases, auto)
next
  case (t_abs L \<Delta> \<Gamma> \<tau>1 \<tau>2 eb e')
  hence Ht: "\<Gamma>;\<Delta> \<turnstile> \<lambda>[\<tau>1].eb : \<tau>1\<rightarrow>\<tau>2" by auto
  have "S;\<lambda>[\<tau>1].eb \<hookrightarrow> S';e'" .
  thus ?case proof cases
    case (e_val Sa v z)
    hence "\<Gamma>,(z:\<tau>1\<rightarrow>\<tau>2);\<Delta> \<Turnstile> OK" using t_abs by (auto simp add: wf_env_cons store_dom_not_in)
    hence "\<Gamma>,(z:\<tau>1\<rightarrow>\<tau>2);\<Delta> \<turnstile> TmFreeVar z : \<tau>1\<rightarrow>\<tau>2" using t_abs by auto
    moreover
    have "\<Turnstile> S,(z\<mapsto>\<lambda>[\<tau>1].eb) : \<Gamma>,(z:\<tau>1\<rightarrow>\<tau>2); \<Delta>"
      using e_val t_abs Ht store_typing_cons[of S \<Gamma> _ "\<lambda>[\<tau>1].eb" "\<tau>1\<rightarrow>\<tau>2" z] by auto
    ultimately
    show ?thesis using e_val by (auto simp add: exI[of _ "{(z\<rhd>\<tau>1\<rightarrow>\<tau>2)}"])
  qed auto
next
  case (t_app \<Delta> \<Gamma> \<tau>1 \<tau>2 e1 e2 e')
  have H1: "\<Gamma>;\<Delta> \<turnstile> e1 : \<tau>1 \<rightarrow> \<tau>2" .
  have H2: "\<Gamma>;\<Delta> \<turnstile> e2 : \<tau>1" .
  have "S;e1 \<cdot> e2 \<hookrightarrow> S';e'" .
  thus ?case using H1 H2 proof cases
    (* propagate 1 *)
    case (e_app_p1 Sa Sa' e1a e1')
    have "\<Delta> \<Turnstile> \<tau>1 \<rightarrow> \<tau>2" using t_app ty_wf_type by auto
    hence "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>';\<Delta> \<and> \<Gamma> \<union> \<Gamma>';\<Delta> \<turnstile> e1' : \<tau>1 \<rightarrow> \<tau>2" using t_app e_app_p1 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>';\<Delta> \<and> \<Gamma> \<union> \<Gamma>';\<Delta> \<turnstile> e1' : \<tau>1 \<rightarrow> \<tau>2" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>';\<Delta> \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>';\<Delta> \<turnstile> e2 : \<tau>1" using t_app by (auto simp add: weakening_tm)
    ultimately
    show ?thesis using e_app_p1 by auto
  next
    (* propagate 2 *)
    case (e_app_p2 Sa Sa' e2a e2' z)
    have "\<Delta> \<Turnstile> \<tau>1" using t_app ty_wf_type by auto
    hence "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>';\<Delta> \<and> \<Gamma> \<union> \<Gamma>';\<Delta> \<turnstile> e2' : \<tau>1" using t_app e_app_p2 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>';\<Delta> \<and> \<Gamma> \<union> \<Gamma>';\<Delta> \<turnstile> e2' : \<tau>1" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>';\<Delta> \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>';\<Delta> \<turnstile> TmFreeVar z : \<tau>1 \<rightarrow> \<tau>2" using t_app e_app_p2 by (auto simp add: weakening)
    ultimately
    show ?thesis using e_app_p2 by auto
  next
    (* beta *)
    case (e_beta Sa \<tau>1_y e1_y y w)
    hence Ht1: "\<Gamma>;\<Delta> \<turnstile> \<lambda>[\<tau>1_y]. e1_y : \<tau>1 \<rightarrow> \<tau>2"
      using store_typing_elim[of Sa \<Gamma> _ _ "\<lambda>[\<tau>1_y].e1_y" "\<tau>1\<rightarrow>\<tau>2"] t_app e_beta by auto
    then obtain L where Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" and Hfin: "finite L"
      and Hs: "\<forall>x. \<not> x free in e1_y \<and> x \<notin> L \<and> \<not> \<Gamma> defines x \<longrightarrow>
                 \<Gamma>,(x:\<tau>1_y);\<Delta> \<turnstile> freshen_tm x e1_y : \<tau>2"
      and Heq: "\<tau>1 = \<tau>1_y" by (rule typing_abs_elim, auto)
    have Ht2: "\<Gamma>;\<Delta> \<turnstile> TmFreeVar y : \<tau>1_y" using e_beta t_app Heq by auto
    obtain z where Hz: "\<not> z free in e1_y \<and> z \<notin> L \<and> \<not> \<Gamma> defines z"
      using Hfin fresh_exists[of "set (free_tm e1_y) \<union> L \<union> dom \<Gamma>"]
        Hgok wf_env_def[of \<Gamma>] dom_finite[of \<Gamma>] by auto
    show ?thesis proof (rule exI[of _ "{}"], auto)
      have "\<Gamma>,(z:\<tau>1_y);\<Delta> \<turnstile> freshen_tm z e1_y : \<tau>2" using Hs Hz by auto
      hence "\<Gamma>;\<Delta> \<turnstile> vsubst_tm (TmFreeVar y) 0 e1_y : \<tau>2"
        using Hz Hz Ht2 by (auto simp add: substitution)
      thus "\<Gamma>;\<Delta> \<turnstile> e' : \<tau>2" using e_beta by auto
    next
      show "\<Turnstile> S' : \<Gamma>; \<Delta>" using t_app e_beta by auto
    qed
  qed auto
next
  case (t_poly L \<Delta> \<Gamma> \<tau> v e')
  have H: "\<forall>x. \<not> x free in term v \<and> \<not> x free in type \<tau> \<and> x \<notin> L \<and> x \<notin> free_ty_env \<Gamma> \<and> x \<notin> dom \<Delta> \<longrightarrow>
    \<Gamma>;\<Delta>,(x:\<star>) \<turnstile> freshen_tmty x v : freshen_ty x \<tau>"
    using prems by blast
  have "S;\<Lambda>[*].v \<hookrightarrow> S';e'" .
  thus ?case proof cases
    case (e_val S va z)
    hence "z \<notin> dom \<Gamma>" using prems by (auto simp add: store_dom_not_in)
    hence Hok: "\<Gamma>,(z:\<forall>.\<tau>);\<Delta> \<Turnstile> OK" using t_poly wf_env_cons by auto
    have "\<Gamma>;\<Delta> \<turnstile> \<Lambda>[*].v : \<forall>.\<tau>" using e_val t_poly by auto
    hence "\<Turnstile> S,(z\<mapsto>\<Lambda>[*].v) : \<Gamma>,(z:\<forall>.\<tau>);\<Delta>" 
      using t_poly e_val store_typing_cons[of S \<Gamma> \<Delta> "\<Lambda>[*].v" "\<forall>.\<tau>" z] by auto
    moreover
    hence "\<Gamma>,(z:\<forall>.\<tau>);\<Delta> \<turnstile> TmFreeVar z : \<forall>.\<tau>" using Hok t_var by auto
    ultimately
    show ?thesis using e_val by (auto simp add:  exI[of _ "{(z\<rhd>\<forall>.\<tau>)}"] exI[of _ "{}"])
  qed auto
next
  case (t_tyapp \<Delta> \<Gamma> \<tau> \<tau>1 e1 e')
  have H: "\<Gamma>;\<Delta> \<turnstile> e1 : \<forall>.\<tau>1" .
  have H': "\<Delta> \<Turnstile> \<forall>.\<tau>1" using prems ty_wf_type by auto
  have H'': "\<Delta> \<Turnstile> \<tau>" using prems by auto
  have "S;e1 \<langle>\<tau>\<rangle> \<hookrightarrow> S';e'" .
  thus ?case using H H' H'' proof cases
    case (e_tyapp_p S S' \<tau> e e1')
    have "\<exists>\<Gamma>'. \<Turnstile> S': \<Gamma>\<union>\<Gamma>' ; \<Delta> \<and> \<Gamma>\<union>\<Gamma>'; \<Delta> \<turnstile> e1' : \<forall>.\<tau>1" using t_tyapp e_tyapp_p H' by auto
    then obtain \<Gamma>' where "\<Turnstile> S': \<Gamma>\<union>\<Gamma>' ; \<Delta> \<and> \<Gamma>\<union>\<Gamma>';\<Delta> \<turnstile> e1' : \<forall>.\<tau>1" by blast
    moreover
    hence "\<Gamma>\<union>\<Gamma>';\<Delta> \<Turnstile> OK" using t_tyapp ty_wf_env by auto
    moreover
    hence "\<Delta> \<Turnstile> TY" using wf_env_wf_env_ty by auto
    hence "\<Delta> \<Turnstile> \<tau>" using t_tyapp e_tyapp_p ty_wf_type by auto
    hence "\<Gamma>\<union>\<Gamma>';\<Delta> \<turnstile> e1' \<langle>\<tau>\<rangle> : vsubst_ty \<tau> 0 \<tau>1" using prems by auto
    ultimately
    show ?thesis using e_tyapp_p by blast
  next
    case (e_tybeta Sa _ v z)
    have "z\<rhd> \<forall>.\<tau>1 \<in> \<Gamma>" using prems by auto
    hence Ht: "\<Gamma>;\<Delta> \<turnstile> \<Lambda>[*]. v : \<forall>.\<tau>1"
      using e_tybeta t_tyapp store_typing_elim[of S \<Gamma> \<Delta> z _ "\<forall>.\<tau>1"] by auto
    then obtain L where Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" and Hfin: "finite L"
      and Hs: "\<forall>a. \<not> a free in term v \<and> \<not> a free in type \<tau>1 \<and> a \<notin> L \<and>
                            \<not> a free in env \<Gamma> \<and> \<not> \<Delta> defines a \<longrightarrow>
                          \<Gamma>;\<Delta>,(a:\<star>) \<turnstile> freshen_tmty a v : freshen_ty a \<tau>1"
      and Hv: "value v"
      by (rule typing_poly_elim, auto)
    then obtain k where Hk: "k \<notin> set (free_tmty v) \<and> k \<notin> set (free_ty \<tau>1) \<and> k \<notin> L \<and>
                             k \<notin> free_ty_env \<Gamma> \<and> k \<notin> Environ.dom \<Delta>"
      using fresh_exists[of "set (free_tmty v) \<union> set (free_ty \<tau>1) \<union> L \<union>
        free_ty_env \<Gamma> \<union> dom \<Delta>"] wf_env_def[of \<Gamma>] wf_env_ty_def dom_finite[of \<Delta>]
        free_env_finite[of \<Gamma>] by auto
    show ?thesis proof (rule exI[of _ "{}"], auto)
      show "\<Turnstile> S' : \<Gamma>; \<Delta>" using e_tybeta t_tyapp by auto
    next
      have "\<Gamma>;\<Delta>,(k:\<star>) \<Turnstile> OK" using ty_wf_env Hs Hk by auto
      hence "\<Gamma> = vsubst_ty_env (TyFreeVar k) 0 \<Gamma>" and Hn1: "\<Gamma> = vsubst_ty_env \<tau> 0 \<Gamma>"
        using wf_env_subst_eq[of \<Gamma> _ \<tau>] wf_env_subst_eq[of \<Gamma> "\<Delta>,(k:\<star>)" "TyFreeVar k"] by auto
      hence  "vsubst_ty_env (TyFreeVar k) 0 \<Gamma>;\<Delta>,(k:\<star>) \<turnstile> freshen_tmty k v : freshen_ty k \<tau>1"
        using Hs Hk by auto
      moreover
      have "vsubst_ty_env \<tau> 0 \<Gamma>;\<Delta> \<Turnstile> OK" using Hn1 using prems by auto
      ultimately
      have "vsubst_ty_env \<tau> 0 \<Gamma>;\<Delta> \<turnstile> vsubst_tmty \<tau> 0 v : vsubst_ty \<tau> 0 \<tau>1"
        using prems substitution_ty[of v _ \<tau> 0 \<Gamma> \<Delta> k \<tau>1] by auto
      thus "\<Gamma>;\<Delta> \<turnstile> e' : vsubst_ty \<tau> 0 \<tau>1" using Hn1 e_tybeta by auto 
    qed
  qed auto
next
  case (t_new \<Delta> \<Gamma> \<tau> e ef')
  have Ht: "\<Gamma>;\<Delta> \<turnstile> e : \<tau>" .
  hence Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" by (auto simp add: ty_wf_env)
  have "S; TmNew e \<hookrightarrow> S'; ef'" .
  thus ?case using Ht proof cases
    (* propagate *)
    case (e_new_p Sa Sa' ea ea')
    hence H: "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>' ; \<Delta> \<and> \<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> ea': \<tau>" using prems by auto
    then obtain \<Gamma>' where Hst: "\<Turnstile> S': \<Gamma> \<union> \<Gamma>' ; \<Delta>" and Ht: "\<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> ea': \<tau>" by auto
    hence "\<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> TmNew ea': TyRef \<tau>" by auto
    thus ?thesis using t_new e_new_p Ht by auto
  next
    (* calculate *)
    case (e_new_c Sa v x z)
    hence "\<not> \<Gamma> defines z" using t_new by (auto simp add: store_dom_not_in)
    moreover have "\<Delta> \<Turnstile> \<tau>" using t_new ty_wf_type by auto
    ultimately have "\<Gamma>,(z:\<tau>);\<Delta> \<Turnstile> OK" using Hgok by (auto simp add: wf_env_cons)
    hence Htz: "\<Gamma>,(z:\<tau>);\<Delta> \<turnstile> TmLoc z : TyRef \<tau>" by auto
    have "S \<Turnstile> Store" using t_new by (auto simp add: store_ty_store)
    moreover
    hence "value v" using e_new_c by (auto simp add: store_ok_is_value)
    moreover
    have "\<Gamma>;\<Delta> \<turnstile> v : \<tau>" using t_new e_new_c by (auto simp add: store_typing_elim)
    ultimately
    have "\<Turnstile> S,(z\<mapsto>v):\<Gamma>,(z:\<tau>);\<Delta>" using t_new e_new_c by (auto simp add: store_typing_cons)
    thus ?thesis using Htz e_new_c by (auto simp add: exI[of _ "{(z\<rhd>\<tau>)}"])
  qed auto
next
  case (t_assign \<Delta> \<Gamma> \<tau> e1 e2 ef')
  have H1: "\<Gamma>;\<Delta> \<turnstile> e1 : TyRef \<tau>" .
  have H2: "\<Gamma>;\<Delta> \<turnstile> e2 : \<tau>" .
  hence Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" by (auto simp add: ty_wf_env)
  hence Hkdok: "\<Delta> \<Turnstile> TY" by (auto simp add: wf_env_wf_env_ty)
  have "S; e1 := e2 \<hookrightarrow> S'; ef'" .
  thus ?case using H1 H2 proof cases
    (* propagate 1 *)
    case (e_assign_p1 Sa Sa' e1a e1')
    have "\<Delta> \<Turnstile> TyRef \<tau>" using t_assign ty_wf_type by auto
    hence "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>'; \<Delta> \<and> \<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> e1' : TyRef \<tau>"
      using t_assign e_assign_p1 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>'; \<Delta> \<and> \<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> e1' : TyRef \<tau>" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>'; \<Delta> \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> e2 : \<tau>" using t_assign by (auto simp add: weakening_tm)
    ultimately
    show ?thesis using e_assign_p1 by auto
  next
    (* propagate 2 *)
    case (e_assign_p2 Sa Sa' e2a e2' z)
    have "\<Delta> \<Turnstile> \<tau>" using t_assign ty_wf_type by auto
    hence "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>'; \<Delta> \<and> \<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> e2' : \<tau>" using t_assign e_assign_p2 by auto
    then obtain \<Gamma>' where "\<Turnstile> S' : \<Gamma> \<union> \<Gamma>'; \<Delta> \<and> \<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> e2' : \<tau>" by auto
    moreover
    hence "\<Gamma> \<union> \<Gamma>'; \<Delta> \<Turnstile> OK" by (auto simp add: ty_wf_env)
    hence "\<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> TmFreeVar z : TyRef \<tau>"
      using t_assign e_assign_p2 by (auto simp add: weakening_tm)
    ultimately
    show ?thesis using e_assign_p2 by auto
  next
    (* calculate *)
    case (e_assign_c Sa l v x y)
    hence Ht1: "\<Gamma>;\<Delta> \<turnstile> TmLoc l : TyRef \<tau>"
      using store_typing_elim[of Sa \<Gamma> _ _ "TmLoc l" "TyRef \<tau>"] t_assign e_assign_c by auto
    hence Hloc: "(l\<rhd>\<tau>) \<in> \<Gamma>" by auto
    have Ht2: "\<Gamma>;\<Delta> \<turnstile> TmFreeVar y : \<tau>" using e_assign_c t_assign by auto
    show ?thesis proof (rule exI[of _ "{}"], auto)
      have "S \<Turnstile> Store" using t_assign by (auto simp add: store_ty_store)
      moreover
      hence "value v" using e_assign_c by (auto simp add: store_ok_is_value)
      moreover
      have "\<Gamma>;\<Delta> \<turnstile> v : \<tau>" using t_assign e_assign_c by (auto simp add: store_typing_elim)
      moreover
      have "\<Gamma>;\<Delta> \<turnstile> TmFreeVar l : \<tau>" using Ht1 by auto
      ultimately
      have "\<Turnstile> update S l v : \<Gamma>; \<Delta>" using t_assign Hloc st_ty_update by auto
      thus "\<Turnstile> S' : \<Gamma>; \<Delta>" using e_assign_c by auto
    next
      show "\<Gamma>;\<Delta> \<turnstile> ef' : TyUnit" using Hgok e_assign_c by auto
    qed
  qed auto
next
  case (t_deref \<Delta> \<Gamma> \<tau> e ef')
  have Ht: "\<Gamma>;\<Delta> \<turnstile> e : TyRef \<tau>" .
  hence Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" by (auto simp add: ty_wf_env)
  have "S; TmDeref e \<hookrightarrow> S'; ef'" .
  thus ?case proof cases
    (* propagate *)
    case (e_deref_p Sa Sa' ea ea')
    have "\<Delta> \<Turnstile> TyRef \<tau>" using t_deref ty_wf_type by auto
    hence H: "\<exists>\<Gamma>'. \<Turnstile> S' : \<Gamma> \<union> \<Gamma>'; \<Delta> \<and> \<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> ea': TyRef \<tau>" using prems by auto
    then obtain \<Gamma>' where Hst: "\<Turnstile> S': \<Gamma> \<union> \<Gamma>'; \<Delta>" and Ht': "\<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> ea': TyRef \<tau>" by auto
    hence "\<Gamma> \<union> \<Gamma>'; \<Delta> \<turnstile> TmDeref ea': \<tau>" by auto
    thus ?thesis using t_deref e_deref_p Ht' by auto
  next
    (* calculate *)
    case (e_deref_c Sa l v z)
    hence "z\<rhd>TyRef \<tau> \<in> \<Gamma>" using t_deref by auto
    hence "\<Gamma>;\<Delta> \<turnstile> TmLoc l : TyRef \<tau>"
      using t_deref e_deref_c store_typing_elim[of S \<Gamma> _ z _ "TyRef \<tau>"] by auto
    hence "\<Gamma>;\<Delta> \<turnstile> v : \<tau>"
      using t_deref e_deref_c store_typing_elim[of S \<Gamma> _ l v \<tau>] by auto
    thus ?thesis using t_deref e_deref_c by (auto simp add: exI[of _ "{}"])
  qed auto
next
  case (t_loc \<Delta> \<Gamma> \<tau> l e')
  have "S;TmLoc l \<hookrightarrow> S';e'" .
  thus ?case proof cases
    case (e_val Sa v z)
    hence "\<Gamma>,(z:TyRef \<tau>);\<Delta> \<Turnstile> OK" using t_loc by (auto simp add: wf_env_cons store_dom_not_in)
    hence "\<Gamma>,(z:TyRef \<tau>);\<Delta> \<turnstile> TmFreeVar z : TyRef \<tau>" using t_loc by auto
    moreover
    have "\<Turnstile> S,(z\<mapsto>TmLoc l) : \<Gamma>,(z:TyRef \<tau>); \<Delta>"
      using t_loc e_val store_typing_cons[of _ _ _ "TmLoc l" "TyRef \<tau>"] by auto
    ultimately
    show ?thesis using e_val by (auto simp add: exI[of _ "{(z\<rhd>TyRef \<tau>)}"])
  qed auto
qed
qed

theorem progress:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  and "\<Turnstile> S : \<Gamma>;\<Delta>"
  shows "not_stuck e S"
proof -
  have Hstok: "S \<Turnstile> Store" using prems by (auto simp add: store_ty_store)
  show ?thesis using prems proof induct
    case (t_unit \<Delta> \<Gamma>)
    obtain z where "\<not> S defines z"
      using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
    hence "S;TmUnit \<hookrightarrow> S,(z\<mapsto>TmUnit);TmFreeVar z" using Hstok by auto
    thus ?case by auto
  next
    case (t_abs L \<Delta> \<Gamma> \<tau>1 \<tau>2 e_b)
    obtain z where "\<not> S defines z"
      using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
    hence "S;\<lambda>[\<tau>1]. e_b \<hookrightarrow> S,(z\<mapsto>\<lambda>[\<tau>1]. e_b);TmFreeVar z" using Hstok by auto
    thus ?case by auto
  next
    case (t_app \<Delta> \<Gamma> \<tau>1 \<tau>2 e1 e2)
    hence "isVar e1 \<or> makes_progress e1 S" by auto
    thus ?case proof
      assume "isVar e1"
      then obtain x where Hx: "e1 = TmFreeVar x" by auto
      have "isVar e2 \<or> makes_progress e2 S" using t_app by auto
      thus ?case proof
        assume "isVar e2"
        then obtain y where Hy: "e2 = TmFreeVar y" by auto
        have "\<Gamma> defines x" using Hx t_app by (auto simp add: dom_def)
        hence "S defines x" using t_app by (auto simp add: store_dom_in)
        then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
          using Hstok store_ok_exists_value by auto
        hence "\<Gamma>;\<Delta> \<turnstile> v : \<tau>1 \<rightarrow> \<tau>2" using t_app Hx by (auto simp add: store_typing_elim)
        then obtain e_b where "v = \<lambda>[\<tau>1]. e_b" using Hv by (cases v, auto)
        hence "S;x\<Down>\<lambda>[\<tau>1]. e_b" using Hxv by auto
        hence "S;TmFreeVar x \<cdot> TmFreeVar y \<hookrightarrow> S;freshen_tm y e_b" ..
        thus ?thesis using Hx Hy by force
      next
        assume "makes_progress e2 S"
        thus ?thesis using Hx by auto
      qed
    next
      assume "makes_progress e1 S"
      thus ?thesis by auto
    qed
  next
    case (t_poly L \<Delta> \<Gamma> \<tau> v)
    obtain z where "\<not> S defines z"
      using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
    hence "S;\<Lambda>[*]. v \<hookrightarrow> S,(z\<mapsto>\<Lambda>[*]. v);TmFreeVar z" using Hstok by auto
    thus ?case by auto
  next
    case (t_tyapp \<Delta> \<Gamma> \<tau> \<tau>1 e)
    hence "isVar e \<or> makes_progress e S" by auto
    thus ?case proof
      assume "makes_progress e S"
      thus ?thesis by auto
    next
      assume "isVar e"
      then obtain x where Hx: "e = TmFreeVar x" by auto
      hence "\<Gamma> defines x" using t_tyapp by (auto simp add: dom_def)
      then obtain v where "(x\<mapsto>v) \<in> S \<and> value v"
        using t_tyapp store_dom_in store_ty_store store_ok_exists_value[of S x] by auto
      moreover
      from store_typing_elim have "\<Gamma>;\<Delta> \<turnstile> v : \<forall>. \<tau>1" using prems by auto
      then obtain w where "v = \<Lambda>[*].w" using prems by (cases v, auto)
      ultimately
      have "S;x\<Down>(\<Lambda>[*].w)" using prems by auto
      hence "S;e\<langle>\<tau>\<rangle> \<hookrightarrow> S; vsubst_tmty \<tau> 0 w" using e_tybeta Hx by auto
      thus ?case using exI[of _ S] exI[of _ "vsubst_tmty \<tau> 0 w"] by auto
    qed
  next
    case (t_new \<Delta> \<Gamma> \<tau> e)
    hence "isVar e \<or> makes_progress e S" by auto
    thus ?case proof
      assume "isVar e"
      then obtain x where Hx: "e = TmFreeVar x" by auto
      have "\<Gamma> defines x" using Hx t_new by (auto simp add: dom_def)
      then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
        using Hstok t_new store_dom_in store_ok_exists_value[of S x] by auto
      obtain z where "\<not> S defines z"
        using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
      hence "S;TmNew (TmFreeVar x) \<hookrightarrow> S,(z\<mapsto>v);TmLoc z" using prems Hx by auto
      thus ?thesis using Hx by force
    next
      assume "makes_progress e S"
      thus ?thesis by auto
    qed
  next
    case (t_assign \<Delta> \<Gamma> \<tau> e1 e2)
    hence "isVar e1 \<or> makes_progress e1 S" by auto
    thus ?case proof
      assume "isVar e1"
      then obtain x where Hx: "e1 = TmFreeVar x" by auto
      have "isVar e2 \<or> makes_progress e2 S" using t_assign by auto
      thus ?case proof
        assume "isVar e2"
        then obtain y where Hy: "e2 = TmFreeVar y" by auto
        hence "\<Gamma> defines y" using t_assign by (auto simp add: dom_def)
        then obtain v' where "(y\<mapsto>v') \<in> S" and "value v'"
          using t_assign Hstok store_dom_in store_ok_exists_value[of S y] by auto
        hence "S;y\<Down>v'" by auto
        moreover
        have "\<Gamma> defines x" using Hx t_assign by (auto simp add: dom_def)
        then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
          using t_assign Hstok store_dom_in store_ok_exists_value[of S x] by auto
        hence "\<Gamma>;\<Delta> \<turnstile> v : TyRef \<tau>" using t_assign Hx by (auto simp add: store_typing_elim)
        then obtain l where "v = TmLoc l" using Hv by (cases v, auto)
        hence "S;x\<Down>TmLoc l" using Hxv by auto
        ultimately
        have "S;TmFreeVar x := TmFreeVar y \<hookrightarrow> update S l v';TmUnit" by auto
        thus ?thesis using Hx Hy by force
      next
        assume "makes_progress e2 S"
        thus ?thesis using Hx by auto
      qed
    next
      assume "makes_progress e1 S"
      thus ?thesis by auto
    qed
  next
    case (t_deref \<Delta> \<Gamma> \<tau> e)
    hence "isVar e \<or> makes_progress e S" by auto
    thus ?case proof
      assume "isVar e"
      then obtain x where Hx: "e = TmFreeVar x" by auto
      have "\<Gamma> defines x" using Hx t_deref by (auto simp add: dom_def)
      then obtain v where Hxv: "(x\<mapsto>v) \<in> S" and Hv: "value v"
        using t_deref Hstok store_dom_in store_ok_exists_value[of S x] by auto
      hence Htv: "\<Gamma>;\<Delta> \<turnstile> v : TyRef \<tau>" using t_deref Hx by (auto simp add: store_typing_elim)
      then obtain l where Heqv: "v = TmLoc l" using Hv by (cases v, auto)
      hence "S;x\<Down>TmLoc l" using Hxv by auto
      moreover
      have "\<Gamma> defines l" using Htv Heqv by (auto simp add: dom_def)
      then obtain v' where "S;l\<Down>v'"
        using t_deref store_dom_in Hstok store_ok_exists_value[of S l] by auto
      ultimately have "S;TmDeref (TmFreeVar x) \<hookrightarrow> S;v'" by auto
      thus ?thesis using Hx by force
    next
      assume "makes_progress e S"
      thus ?thesis by auto
    qed
  next
    case (t_loc \<Delta> \<Gamma> \<tau> l)
    obtain z where "\<not> S defines z"
      using prems fresh_exists[of "dom S"] by (auto simp add: dom_finite store_ty_finite)
    hence "S;TmLoc l \<hookrightarrow> S,(z\<mapsto>TmLoc l);TmFreeVar z" using Hstok by auto
    thus ?case by auto
  qed auto
qed

theorem safety:
  assumes Ht: "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  and Hsem: "S;e \<hookrightarrow>* S';e'"
  and Hstok: "\<Turnstile> S : \<Gamma>; \<Delta>"
  shows "not_stuck e' S'"
using Hsem Ht Hstok proof (induct arbitrary: \<Gamma>)
  case e_refl thus ?case using progress by auto
next
  case e_trans thus ?case using progress preservation by blast
qed

theorem program_safety:
  assumes "{};{} \<turnstile> e : \<tau>"
  and "{};e \<hookrightarrow>* S';e'"
  shows "not_stuck e' S'"
using prems by (rule safety, auto simp add: storeTyping_def wf_store_def)

end
