theory Semantics
imports Syntax

begin

datatype StTerm = StTerm "tm"
types store = "(name * StTerm) set"

syntax
  var_stterm :: "name \<Rightarrow> tm \<Rightarrow> (name * StTerm)"  ("(_\<mapsto>_)" [90,90] 90)
translations
  "(x\<mapsto>v)" \<rightleftharpoons> "(x, StTerm v)"

syntax
  addmapping :: "store \<Rightarrow> name \<Rightarrow> tm \<Rightarrow> store"      ("_','(_\<mapsto>_')" [90,0,0] 91)
translations
  "\<delta>,(x\<mapsto>v)" \<rightleftharpoons> "insert (x\<mapsto>v) \<delta>"

consts Var :: "tm set"
syntax
  "_Var" :: "tm \<Rightarrow> bool"   ("isVar")
translations
  "isVar e" \<rightleftharpoons> "e \<in> Var"

inductive Var
  intros
  v_var[intro!]  : "isVar (TmFreeVar x)"

inductive_cases isVar_var_elim[elim!]: "isVar e"

constdefs
  wf_store :: "store \<Rightarrow> bool" ("_ \<Turnstile> Store" [50] 50)
  "wf_store S \<equiv> finite S \<and> (\<forall>x. \<forall>v1. \<forall>v2. (x\<mapsto>v1) \<in> S \<and> (x\<mapsto>v2) \<in> S \<longrightarrow> v1 = v2 \<and> value v1)"

lemma wf_store_cons:
  assumes "S \<Turnstile> Store"
  and "\<not> S defines z"
  and "value v"
  shows "S,(z\<mapsto>v) \<Turnstile> Store"
using prems by (auto simp add: dom_def wf_store_def)

consts ValComp :: "store \<Rightarrow> name \<Rightarrow> tm \<Rightarrow> bool"   ("_;_\<Down>_" [50,80,80] 50)
defs ValCom[simp]:
  "S;z\<Down>v \<equiv> (z\<mapsto>v) \<in> S"

constdefs update :: "store \<Rightarrow> name \<Rightarrow> tm \<Rightarrow> store"
  "update S x v \<equiv> { p | p. p \<in> S \<and> fst p \<noteq> x},(x\<mapsto>v)"

lemma store_update_ok:
  assumes "S \<Turnstile> Store"
  and "value v"
  shows "update S z v \<Turnstile> Store"
using prems by (auto simp add: wf_store_def update_def)

consts Eval :: "(store * tm * store * tm) set"
syntax
  "_EVal" :: "store \<Rightarrow> tm \<Rightarrow> store \<Rightarrow> tm \<Rightarrow> bool"  ("_';_\<hookrightarrow>_';_" [50,80,50,80] 50)
translations
  "S1;e1\<hookrightarrow>S2;e2" \<rightleftharpoons> "(S1, e1, S2, e2) \<in> Eval"

inductive Eval
  intros
  e_val[intro!]       : "\<lbrakk> \<not> S defines z; S \<Turnstile> Store; value v \<rbrakk> \<Longrightarrow> S;v \<hookrightarrow> S,(z\<mapsto>v);TmFreeVar z"
  e_app_p1[intro!]    : "\<lbrakk> S;e1\<hookrightarrow>S';e1' \<rbrakk> \<Longrightarrow> S;e1\<cdot>e2 \<hookrightarrow> S';e1'\<cdot>e2"
  e_app_p2[intro!]    : "\<lbrakk> S;e2\<hookrightarrow>S';e2' \<rbrakk> \<Longrightarrow> S;(TmFreeVar x)\<cdot>e2 \<hookrightarrow> S';(TmFreeVar x)\<cdot>e2'"
  e_beta[intro!]      : "\<lbrakk> S;z\<Down>\<lambda>[\<tau>].e1 \<rbrakk> \<Longrightarrow>
                           S;(TmFreeVar z)\<cdot>(TmFreeVar y) \<hookrightarrow> S;freshen_tm y e1"
  e_tyapp_p[intro!]   : "\<lbrakk> S;e\<hookrightarrow>S';e' \<rbrakk> \<Longrightarrow> S;e\<langle>\<tau>\<rangle> \<hookrightarrow> S';e'\<langle>\<tau>\<rangle>"
  e_tybeta[intro!]    : "\<lbrakk> S;z\<Down>\<Lambda>[*].e \<rbrakk> \<Longrightarrow> S;(TmFreeVar z)\<langle>\<tau>\<rangle> \<hookrightarrow> S;vsubst_tmty \<tau> 0 e"
  e_new_p[intro!]     : "\<lbrakk> S;e\<hookrightarrow>S';e' \<rbrakk> \<Longrightarrow> S;TmNew e \<hookrightarrow> S';TmNew e'"
  e_new_c[intro!]     : "\<lbrakk> \<not> S defines z; S;x\<Down>v \<rbrakk> \<Longrightarrow> S;TmNew (TmFreeVar x) \<hookrightarrow> S,(z\<mapsto>v);TmLoc z"
  e_assign_p1[intro!] : "\<lbrakk> S;e1\<hookrightarrow>S';e1' \<rbrakk> \<Longrightarrow> S;TmAssign e1 e2 \<hookrightarrow> S';TmAssign e1' e2"
  e_assign_p2[intro!] : "\<lbrakk> S;e2\<hookrightarrow>S';e2' \<rbrakk> \<Longrightarrow>
                           S;TmAssign (TmFreeVar x) e2 \<hookrightarrow> S';TmAssign (TmFreeVar x) e2'"
  e_assign_c[intro!]  : "\<lbrakk> S;x\<Down>TmLoc l; S;y\<Down>v \<rbrakk> \<Longrightarrow> 
                           S;TmAssign (TmFreeVar x) (TmFreeVar y) \<hookrightarrow> update S l v; TmUnit"
  e_deref_p[intro!]   : "\<lbrakk> S;e\<hookrightarrow>S';e' \<rbrakk> \<Longrightarrow> S;TmDeref e \<hookrightarrow> S';TmDeref e'"
  e_deref_c[intro!]   : "\<lbrakk> S;x\<Down>TmLoc l; S;l\<Down>v \<rbrakk> \<Longrightarrow> S;TmDeref (TmFreeVar x) \<hookrightarrow> S;v"

inductive_cases e_eval_unit[elim!]:      "S;TmUnit \<hookrightarrow> S';e'" 
inductive_cases e_eval_var[elim!]:       "S;TmVar n \<hookrightarrow> S';e'" 
inductive_cases e_eval_freevar[elim!]:   "S;TmFreeVar n \<hookrightarrow> S';e'" 
inductive_cases e_eval_app_p1[elim!]:    "S;e1 \<cdot> e2 \<hookrightarrow> S';e'" 
inductive_cases e_eval_app_p2[elim!]:    "S;(TmFeeVar x) \<cdot> e2 \<hookrightarrow> S';e'" 
inductive_cases e_eval_app_beta[elim!]:  "S;(TmFeeVar x) \<cdot> (TmFreeVar y) \<hookrightarrow> S';e'" 
inductive_cases e_eval_poly[elim!]:      "S;\<Lambda>[*].e \<hookrightarrow> S';e'"
inductive_cases e_eval_tapp_beta[elim!]: "S;(TmFreeVar x) \<langle>\<tau>\<rangle> \<hookrightarrow> S';e'" 
inductive_cases e_eval_new_p[elim!]:     "S;TmNew e \<hookrightarrow> S';e'" 
inductive_cases e_eval_new_c[elim!]:     "S;TmNew (TmFreeVar x) \<hookrightarrow> S';e'" 
inductive_cases e_eval_assign_p1[elim!]: "S;e1 := e2 \<hookrightarrow> S';e'" 
inductive_cases e_eval_assign_p2[elim!]: "S;(TmFreeVar x) := e2 \<hookrightarrow> S';e'" 
inductive_cases e_eval_assign_c[elim!]:  "S;(TmFreeVar x) := (TmFreeVar y) \<hookrightarrow> S';e'" 
inductive_cases e_deref_p[elim!]:        "S;TmDeref e \<hookrightarrow> S';e'" 
inductive_cases e_deref_c[elim!]:        "S;TmDeref (TmFreeVar x) \<hookrightarrow> S';e'" 
inductive_cases e_loc[elim!]:            "S;TmLoc n \<hookrightarrow> S';e'" 

consts makes_progress :: "tm \<Rightarrow> store \<Rightarrow> bool"
defs makes_progress[simp]: "makes_progress e S \<equiv> \<exists>e'. \<exists>S'. S;e \<hookrightarrow> S';e'"

consts not_stuck :: "tm \<Rightarrow> store \<Rightarrow> bool"
defs not_stuck[simp]: "not_stuck e S \<equiv> isVar e \<or> makes_progress e S"

consts EvalMany :: "(store * tm * store * tm) set"
syntax
  "_EValM" :: "store \<Rightarrow> tm \<Rightarrow> store => tm \<Rightarrow> bool"   ("_';_\<hookrightarrow>*_';_" [50,80,50,80] 50)
translations
  "S1;e1 \<hookrightarrow>* S2;e2" \<rightleftharpoons> "(S1, e1, S2, e2) \<in> EvalMany"

inductive EvalMany
  intros
  e_refl[intro!]: "S;e \<hookrightarrow>* S;e"
  e_trans[intro!]: "\<lbrakk> S;e \<hookrightarrow> S';e'; S';e' \<hookrightarrow>* S'';e'' \<rbrakk> \<Longrightarrow> S;e \<hookrightarrow>* S'';e''"

lemma store_ok_is_value:
  assumes "S \<Turnstile> Store"
  and "(x\<mapsto>v) \<in> S"
  shows "value v"
using prems by (auto simp add: wf_store_def)

lemma store_ok_exists_value:
  assumes "S \<Turnstile> Store"
  and "x \<in> dom S"
  shows "\<exists>v. (x\<mapsto>v) \<in> S \<and> value v"
using prems proof (auto simp add: dom_def store_ok_is_value)
  fix b
  assume "S \<Turnstile> Store"
    and "(x, b) \<in> S"
  show "\<exists>v. x\<mapsto>v \<in> S \<and> value v" using prems proof (cases b)
    fix tm
    assume "(x, b) \<in> S"
      and "b = StTerm tm"
    have "(x\<mapsto>tm) \<in> S" using prems by auto
    hence "value tm" using prems store_ok_is_value by auto
    thus "\<exists>v. x\<mapsto>v \<in> S \<and> value v" using prems by blast
  qed
qed

end
